@@ -14,8 +14,9 @@ use Lemonldap::NG::Portal::Main::Constants qw(
   PE_FORMEMPTY
   PE_SENDRESPONSE
 );
+use Lemonldap::NG::Common::Util qw/display2F/;
 
-our $VERSION = '2.0.16';
+our $VERSION = '2.19.0';
 
 extends qw(
   Lemonldap::NG::Portal::Main::SecondFactor
@@ -46,14 +47,12 @@ sub run {
     $self->logger->debug( $self->prefix . '2f: generate form' );
 
     # Prepare form
-    my ( $checkLogins, $stayConnected ) = $self->getFormParams($req);
     my $tmp = $self->p->sendHtml(
         $req,
         'totp2fcheck',
         params => {
-            TOKEN         => $token,
-            CHECKLOGINS   => $checkLogins,
-            STAYCONNECTED => $stayConnected
+            TOKEN => $token,
+            $self->get2fTplParams($req),
         }
     );
     $self->logger->debug( $self->prefix . '2f: prepare verification' );
@@ -65,6 +64,7 @@ sub run {
 sub verify {
     my ( $self, $req, $session ) = @_;
     my ( $code, $secret, @totp2f );
+    my $uid = $session->{ $self->conf->{whatToTrace} };
     $self->logger->debug( $self->prefix . '2f: verification' );
 
     unless ( $code = $req->param('code') ) {
@@ -73,30 +73,34 @@ sub verify {
     }
 
     @totp2f = $self->find2fDevicesByType( $req, $session, $self->type );
-    $secret = $_->{_secret} foreach @totp2f;
-    unless ($secret) {
-        $self->logger->debug( $self->prefix . '2f: no secret found' );
-        return PE_BADOTP;
+
+    foreach my $device (@totp2f) {
+        $self->logger->debug( "Trying TOTP device " . display2F($device) );
+
+        if ( my $secret = $device->{_secret} ) {
+            my ( $r, $range ) = $self->verifyCode(
+                $self->conf->{totp2fInterval},
+                $self->conf->{totp2fRange},
+                $self->conf->{totp2fDigits},
+                $secret, $code
+            );
+            if ( $r == 1 ) {
+                $req->data->{_2fDevice}  = $device;
+                $req->data->{_2fLogInfo} = { range => $range };
+                return PE_OK;
+            }
+        }
+        else {
+            $self->logger->warn( "TOTP device "
+                  . display2F($device)
+                  . " has no secret for user $uid" );
+        }
     }
 
-    my $r = $self->verifyCode(
-        $self->conf->{totp2fInterval},
-        $self->conf->{totp2fRange},
-        $self->conf->{totp2fDigits},
-        $secret, $code
-    );
-    return PE_ERROR if $r == -1;
-
-    if ($r) {
-        $self->userLogger->info( $self->prefix . '2f: succeed' );
-        return PE_OK;
-    }
-    else {
-        $self->userLogger->notice( $self->prefix
-              . '2f: invalid attempt for '
-              . $session->{ $self->conf->{whatToTrace} } );
-        return PE_BADOTP;
-    }
+    $self->userLogger->notice( $self->prefix
+          . '2f: code did not match any of the registered TOTP for '
+          . $uid );
+    return PE_BADOTP;
 }
 
 1;
