@@ -5,17 +5,22 @@ use Mouse;
 use URI;
 use Lemonldap::NG::Common::FormEncode;
 use Lemonldap::NG::Portal::Main::Constants qw(
-  PE_CAS_SERVICE_NOT_ALLOWED
+  portalConsts
+  URIRE
+  PE_OK
   PE_INFO
   PE_ERROR
   PE_LOGOUT_OK
-  PE_OK
-  PE_BADURL
   PE_SENDRESPONSE
-  URIRE
+  PE_UNKNOWNPARTNER
+  PE_UNAUTHORIZEDURL
+  PE_UNAUTHORIZEDPARTNER
+  PE_BADURL
+  PE_CAS_SERVICE_NOT_ALLOWED
 );
+use URI;
 
-our $VERSION = '2.0.12';
+our $VERSION = '2.19.0';
 
 extends 'Lemonldap::NG::Portal::Main::Issuer',
   'Lemonldap::NG::Portal::Lib::CAS';
@@ -47,7 +52,9 @@ sub init {
     return 0 unless ( $self->loadApp );
     $self->addUnauthRoute(
         ( $self->path ) => {
+            samlValidate    => 'samlNotImplemented',
             serviceValidate => 'serviceValidate',
+            relayLogout     => 'relayLogout',
             validate        => 'validate',
             proxyValidate   => 'proxyValidate',
             proxy           => 'proxy',
@@ -59,10 +66,21 @@ sub init {
         ['GET']
     );
 
+    $self->addUnauthRoute(
+        ( $self->path ) => {
+            samlValidate => 'samlNotImplemented',
+        },
+        ['POST']
+    );
+
     # Add CAS Services, so we can check service= parameter on logout
     foreach my $casSrv ( keys %{ $self->casAppList } ) {
-        if ( my $serviceUrl =
-            $self->casAppList->{$casSrv}->{casAppMetaDataOptionsService} )
+        for my $serviceUrl (
+            split(
+                /\s+/,
+                $self->casAppList->{$casSrv}->{casAppMetaDataOptionsService}
+            )
+          )
         {
             push @{ $self->p->{additionalTrustedDomains} }, $serviceUrl;
             $self->logger->debug(
@@ -78,20 +96,28 @@ sub storeEnvAndCheckGateway {
     my ( $self, $req ) = @_;
     my $service = $self->p->getHiddenFormValue( $req, 'service' )
       || $req->param('service');
-    $service = '' if ( $self->p->checkXSSAttack( 'service', $service ) );
+    $service = '' if ( $self->p->checkXSSAttackUrldc( 'service', $service ) );
     my $gateway = $self->p->getHiddenFormValue( $req, 'gateway' )
       || $req->param('gateway');
 
     if ( $gateway and $gateway eq "true" ) {
-        $self->logger->debug(
-            "Gateway mode requested, redirect without authentication");
-        $req->response( [ 302, [ Location => $service ], [] ] );
-        for my $s ( $self->ipath, $self->ipath . 'Path' ) {
-            $self->logger->debug("Removing $s from pdata")
-              if delete $req->pdata->{$s};
-        }
+        if ( $self->_gatewayAllowedRedirect( $req, $service ) ) {
+            $self->logger->debug(
+                "Gateway mode requested, redirect without authentication");
+            $req->response(
+                [ 302, [ Location => URI->new($service)->as_string ], [] ] );
+            for my $s ( $self->ipath, $self->ipath . 'Path' ) {
+                $self->logger->debug("Removing $s from pdata")
+                  if delete $req->pdata->{$s};
+            }
 
-        return PE_SENDRESPONSE;
+            return PE_SENDRESPONSE;
+        }
+        else {
+            $self->logger->error(
+                "Disallowing redirection to unknown service $service");
+            return PE_UNKNOWNPARTNER;
+        }
     }
 
     if ( $service and $service =~ URIRE ) {
@@ -112,6 +138,24 @@ sub storeEnvAndCheckGateway {
     return PE_OK;
 }
 
+sub _gatewayAllowedRedirect {
+    my ( $self, $req, $service ) = @_;
+
+    my $app                    = $self->getCasApp($service);
+    my $casAccessControlPolicy = $self->conf->{casAccessControlPolicy};
+
+    # Redirect is allowed if there is no access control or if
+    # the service is declared in CAS apps
+    if ( $casAccessControlPolicy !~ /^(error|faketicket)$/i ) {
+        return 1;
+    }
+    if ($app) {
+        return 1;
+    }
+
+    return 0;
+}
+
 # Main method (launched only for authenticated users, see Main/Issuer)
 sub run {
     my ( $self, $req, $target ) = @_;
@@ -138,9 +182,6 @@ sub run {
     # Session ID
     my $session_id = $req->{sessionInfo}->{_session_id} || $req->id;
 
-    # Session creation timestamp
-    my $time = $req->{sessionInfo}->{_utime} || time();
-
     # 1. LOGIN
     if ( $target eq $cas_login ) {
 
@@ -155,15 +196,17 @@ sub run {
         }
 
         my $h = $self->p->processHook( $req, 'casGotRequest', $cas_request );
-        return $h if ( $h != PE_OK );
+        return $self->_failLogin(
+            $req,
+            msg => "casGotRequest hook failed",
+            res => $h
+        ) if ( $h != PE_OK );
 
         my $service = $cas_request->{service};
-        $service = '' if ( $self->p->checkXSSAttack( 'service', $service ) );
+        $service = '' if ( $self->p->checkXSSAttackUrldc( 'service', $service ) );
         my $renew   = $cas_request->{renew};
         my $gateway = $cas_request->{gateway};
 
-        my $casServiceTicket;
-
         # If no service defined, exit
         unless ( defined $service ) {
             $self->logger->debug("No service defined in CAS URL");
@@ -171,10 +214,12 @@ sub run {
         }
 
         unless ( $service =~ m#^(https?://[^/]+)(/.*)?$# ) {
-            $self->logger->error("Bad service $service");
-            return PE_ERROR;
+            return $self->_failLogin(
+                $req,
+                msg => "Bad service $service",
+                res => PE_BADURL,
+            );
         }
-        my ( $host, $uri ) = ( $1, $2 );
         my $app = $self->getCasApp($service);
 
         my $spAuthnLevel = 0;
@@ -207,8 +252,11 @@ sub run {
                 "CAS access control requested on service $service");
 
             unless ($app) {
-                $self->userLogger->error('CAS service not configured');
-                return PE_CAS_SERVICE_NOT_ALLOWED;
+                return $self->_failLogin(
+                    $req,
+                    msg => 'CAS service not configured',
+                    res => PE_UNKNOWNPARTNER,
+                );
             }
 
             # Check if we have sufficient auth level
@@ -230,95 +278,83 @@ sub run {
                     $self->logger->debug("CAS service $service access allowed");
                 }
                 else {
-                    $self->userLogger->warn( 'User '
-                          . $req->sessionInfo->{ $self->conf->{whatToTrace} }
-                          . " is not authorized to access to $app" );
-
-                    if ( $casAccessControlPolicy =~ /^(error)$/i ) {
-                        $self->logger->debug(
-"Return error instead of redirecting user on CAS service"
-                        );
-                        return PE_CAS_SERVICE_NOT_ALLOWED;
-                    }
-
-                    else {
-                        $self->logger->debug(
-                            "Redirect user on CAS service with a fake ticket");
-                        $casServiceTicket = "ST-F4K3T1CK3T";
-                    }
+                    return $self->_failLoginRule( $req, $cas_request, $app );
                 }
             }
         }
 
-        if ($app) {
-            $self->userLogger->notice( 'User '
-                  . $req->sessionInfo->{ $self->conf->{whatToTrace} }
-                  . " is authorized to access to $app" );
-        }
-        else {
-            $self->userLogger->notice( 'User '
-                  . $req->sessionInfo->{ $self->conf->{whatToTrace} }
-                  . " is redirected to $service" );
-        }
-
-        unless ($casServiceTicket) {
-
-            # Check last authentication time to decide if
-            # the authentication is recent or not
-            my $casRenewFlag     = 0;
-            my $last_authn_utime = $req->{sessionInfo}->{_lastAuthnUTime} || 0;
-            if (
-                time() - $last_authn_utime <
-                $self->conf->{portalForceAuthnInterval} )
-            {
-                $self->logger->debug(
-                    "Authentication is recent, will set CAS renew flag to true"
-                );
-                $casRenewFlag = 1;
-            }
-
-            # Create a service ticket
+        # Check last authentication time to decide if
+        # the authentication is recent or not
+        my $casRenewFlag     = 0;
+        my $last_authn_utime = $req->{sessionInfo}->{_lastAuthnUTime} || 0;
+        if (
+            time() - $last_authn_utime <
+            $self->conf->{portalForceAuthnInterval} )
+        {
             $self->logger->debug(
-                "Create a CAS service ticket for service $service");
-
-            my $Sinfos;
-            $Sinfos->{type}    = 'casService';
-            $Sinfos->{service} = $service;
-            $Sinfos->{renew}   = $casRenewFlag;
-            $Sinfos->{_cas_id} = $session_id;
-            $Sinfos->{_utime}  = $time;
-            $Sinfos->{_casApp} = $app;
-
-            my $h = $self->p->processHook( $req, 'casGenerateServiceTicket',
-                $cas_request, $app, $Sinfos );
-            return $h if ( $h != PE_OK );
-
-            my $casServiceSession = $self->getCasSession( undef, $Sinfos );
-
-            unless ($casServiceSession) {
-                $self->logger->error("Unable to create CAS session");
-                return PE_ERROR;
-            }
-
-            my $casServiceSessionID = $casServiceSession->id;
-            $casServiceTicket = "ST-" . $casServiceSessionID;
-
-            $self->logger->debug(
-                "CAS service session $casServiceSessionID created");
+                "Authentication is recent, will set CAS renew flag to true");
+            $casRenewFlag = 1;
         }
 
-        # Redirect to service
-        # cas_request may have been modified by hook
-        my $service_url = $cas_request->{service};
-        $service_url .= ( $service_url =~ /\?/ ? '&' : '?' )
-          . build_urlencoded( ticket => $casServiceTicket );
+        # Create a service ticket
+        $self->logger->debug(
+            "Create a CAS service ticket for service $service");
 
-        $self->logger->debug("Redirect user to $service_url");
+        my $_utime =
+          $self->conf->{casTicketExpiration}
+          ? (
+            time + $self->conf->{casTicketExpiration} - $self->conf->{timeout} )
+          : ( $req->{sessionInfo}->{_utime} || time() );
 
-        $req->{urldc} = $service_url;
+        my $Sinfos;
+        $Sinfos->{type}    = 'casService';
+        $Sinfos->{service} = $service;
+        $Sinfos->{renew}   = $casRenewFlag;
+        $Sinfos->{_cas_id} = $session_id;
+        $Sinfos->{_utime}  = $_utime;
+        $Sinfos->{_casApp} = $app;
 
-        $req->steps( [] );
-        return PE_OK;
+        $h = $self->p->processHook( $req, 'casGenerateServiceTicket',
+            $cas_request, $app, $Sinfos );
+        return $self->_failLogin(
+            $req,
+            msg => 'casGenerateServiceTicket hook failed',
+            res => $h,
+            ( $app ? ( app => $app ) : () ),
+        ) if ( $h != PE_OK );
+
+        my $casServiceSession = $self->getCasSession( undef, $Sinfos );
+
+        unless ($casServiceSession) {
+            return $self->_failLogin(
+                $req,
+                msg => "Unable to create CAS session",
+                ( $app ? ( app => $app ) : () ),
+            );
+        }
+
+        my $casServiceSessionID = $casServiceSession->id;
+        my $casServiceTicket    = "ST-" . $casServiceSessionID;
+
+        $self->logger->debug(
+            "CAS service session $casServiceSessionID created");
+
+        my $user = $req->sessionInfo->{ $self->conf->{whatToTrace} };
+        $self->auditLog(
+            $req,
+            code => "ISSUER_CAS_LOGIN_SUCCESS",
+            ( $app ? ( app => $app ) : () ),
+            service => $service,
+            user    => $user,
+            message => (
+                $app
+                ? ("User $user is authorized to access to $app")
+                : ("User $user is redirected to $service")
+            ),
+        );
+
+        return $self->_redirectUser( $req, $cas_request->{service},
+            $casServiceTicket );
     }
 
     # 2. LOGOUT
@@ -334,7 +370,7 @@ sub run {
         my $logout_url     = $req->param('url');        # CAS 2.0
         my $logout_service = $req->param('service');    # CAS 3.0
         $logout_service = ''
-          if ( $self->p->checkXSSAttack( 'service', $logout_service ) );
+          if ( $self->p->checkXSSAttackUrldc( 'service', $logout_service ) );
 
         # If we use access control, check that the service URL is trusted
         if ( $self->conf->{casAccessControlPolicy} =~ /^(error|faketicket)$/i )
@@ -345,15 +381,17 @@ sub run {
                 $self->userLogger->error(
                         "Untrusted service URL $logout_service"
                       . "specified for CAS Logout" );
-                return PE_BADURL;
+                return PE_UNAUTHORIZEDURL;
             }
         }
 
-        # Delete linked CAS sessions
-        $self->deleteCasSecondarySessions($session_id);
-
         # Delete local session
         if ( my $session = $self->p->getApacheSession($session_id) ) {
+
+           # This will call Issuer logout methods, incluing our own
+           # TODO: call authLogout and deleteSession instead of duplicating code
+            my $res = $self->p->do( $req, [ @{ $self->p->beforeLogout } ] );
+
             unless ( $self->p->_deleteSession( $req, $session ) ) {
                 $self->logger->error("Fail to delete session $session_id ");
             }
@@ -377,11 +415,7 @@ sub run {
             }
 
             if ($logout_service) {
-                $self->logger->debug(
-                    "User will be redirected to $logout_service");
-                $req->{urldc} = $logout_service;
-                $req->steps( [] );
-                return PE_OK;
+                return $self->_redirectUser( $req, $logout_service );
             }
         }
         else {
@@ -455,6 +489,8 @@ sub logout {
     # Delete linked CAS sessions
     $self->deleteCasSecondarySessions($session_id) if ($session_id);
 
+    $self->_send_back_channel_LogoutRequests( $req, $req->sessionInfo );
+
     return PE_OK;
 }
 
@@ -472,23 +508,35 @@ sub validate {
 
     # Required parameters: service and ticket
     unless ( $service and $ticket ) {
-        $self->logger->error("Service and Ticket parameters required");
-        return $self->returnCasValidateError();
+        return $self->_failValidate( $req,
+            msg => "Service and Ticket parameters required" );
     }
 
     $self->logger->debug(
         "Get validate request with ticket $ticket for service $service");
 
     unless ( $ticket =~ s/^ST-// ) {
-        $self->logger->error("Provided ticket is not a service ticket (ST)");
-        return $self->returnCasValidateError();
+        return $self->_failValidate( $req,
+            msg => "Provided ticket is not a service ticket (ST)" );
     }
 
     my $casServiceSession = $self->getCasSession($ticket);
 
     unless ($casServiceSession) {
-        $self->logger->error("Service ticket session $ticket not found");
-        return $self->returnCasValidateError();
+        return $self->_failValidate( $req,
+            msg => "Service ticket session $ticket not found" );
+    }
+
+    # Make sure the token is still valid, we already compensated for
+    # different TTLs when storing _utime
+    if ( $casServiceSession->{data}->{_utime} ) {
+        if (
+            time >
+            ( $casServiceSession->{data}->{_utime} + $self->conf->{timeout} ) )
+        {
+            return $self->_failValidate( $req,
+                msg => "Session $ticket has expired" );
+        }
     }
 
     $self->logger->debug("Service ticket session $ticket found");
@@ -503,17 +551,17 @@ sub validate {
         if (   $service1_uri->rel($service2_uri) eq "./"
             or $service2_uri->rel($service1_uri) eq "./" )
         {
-            $self->logger->notice(
-"Submitted service $service1_uri does not exactly match initial service "
+            $self->logger->notice( "Submitted service $service1_uri"
+                  . " does not exactly match initial service "
                   . $service2_uri
                   . ' but difference is tolerated.' );
         }
         else {
-            $self->logger->error(
-                "Submitted service $service does not match initial service "
-                  . $casServiceSession->data->{service} );
             $self->deleteCasSession($casServiceSession);
-            return $self->returnCasValidateError();
+            return $self->_failValidate( $req,
+                msg =>
+                  "Submitted service $service does not match initial service "
+                  . $casServiceSession->data->{service} );
         }
     }
     else {
@@ -527,11 +575,10 @@ sub validate {
         $self->logger->debug("Renew flag detected ");
 
         unless ( $casServiceSession->data->{renew} ) {
-            $self->logger->error(
-"Authentication renew requested, but not done in former authentication process"
-            );
             $self->deleteCasSession($casServiceSession);
-            return $self->returnCasValidateError();
+            return $self->_failValidate( $req,
+                msg => "Authentication renew requested,"
+                  . " but not done in former authentication process" );
         }
     }
 
@@ -540,11 +587,11 @@ sub validate {
       $self->p->getApacheSession( $casServiceSession->data->{_cas_id} );
 
     unless ($localSession) {
-        $self->logger->warn( "Local session "
+        $self->deleteCasSession($casServiceSession);
+        return $self->_failValidate( $req,
+                msg => "Local session "
               . $casServiceSession->data->{_cas_id}
               . " notfound" );
-        $self->deleteCasSession($casServiceSession);
-        return $self->returnCasValidateError();
     }
 
     # Get username
@@ -556,9 +603,34 @@ sub validate {
     # Return success message
     $self->deleteCasSession($casServiceSession);
 
+    if ( $self->isLogoutEnabled($app) ) {
+        $self->_save_ticket_for_single_logout( $localSession, $service,
+            $req->param('ticket') );
+    }
+
     my $h =
       $self->p->processHook( $req, 'casGenerateValidateResponse', $username );
-    return $self->returnCasValidateError() if ( $h != PE_OK );
+    return $self->_failValidate(
+        $req,
+        res => $h,
+        msg => "casGenerateValidateResponse hook failed",
+        app => $app
+    ) if ( $h != PE_OK );
+
+    my $user   = $localSession->data->{ $self->conf->{whatToTrace} };
+    my $by_app = $app ? " by $app" : "";
+
+    $req->sessionInfo( $localSession->data );
+
+    $self->auditLog(
+        $req,
+        code    => "ISSUER_CAS_VALIDATE_SUCCESS",
+        message =>
+          ("CAS validation succeeded for ${user}${by_app} as $username"),
+        ( $app ? ( app => $app ) : () ),
+        cas_user => $username,
+        user     => $user,
+    );
 
     return $self->returnCasValidateSuccess( $req, $username );
 }
@@ -612,11 +684,16 @@ sub proxy {
             'Error in proxy session management' );
     }
 
+    my $_utime =
+      $self->conf->{casTicketExpiration}
+      ? ( time + $self->conf->{casTicketExpiration} - $self->conf->{timeout} )
+      : $casProxyGrantingSession->data->{_utime};
+
     my $Pinfos;
     $Pinfos->{type}    = 'casProxy';
     $Pinfos->{service} = $targetService;
     $Pinfos->{_cas_id} = $casProxyGrantingSession->data->{_cas_id};
-    $Pinfos->{_utime}  = $casProxyGrantingSession->data->{_utime};
+    $Pinfos->{_utime}  = $_utime;
     $Pinfos->{proxies} = $casProxyGrantingSession->data->{proxies};
 
     $casProxySession->update($Pinfos);
@@ -634,6 +711,16 @@ sub serviceValidate {
     return $self->_validate2( 'SERVICE', $req );
 }
 
+sub samlNotImplemented {
+    my ( $self, $req ) = @_;
+    return $self->p->sendError(
+        $req,
+        "The CAS samlValidate endpoint is not implemented. "
+          . "Please use serviceValidate instead",
+        501
+    );
+}
+
 sub proxyValidate {
     my ( $self, $req ) = @_;
     return $self->_validate2( 'PROXY', $req );
@@ -658,8 +745,11 @@ sub _validate2 {
     # Required parameters: service and ticket
     unless ( $service and $ticket ) {
         $self->logger->error("Service and Ticket parameters required");
-        return $self->returnCasServiceValidateError( $req, 'INVALID_REQUEST',
-            'Missing mandatory parameters (service, ticket)' );
+        return $self->_failValidate2(
+            $req,
+            code => 'INVALID_REQUEST',
+            msg  => 'Missing mandatory parameters (service, ticket)'
+        );
     }
 
     $self->logger->debug( "Get "
@@ -669,23 +759,49 @@ sub _validate2 {
     # Get CAS session corresponding to ticket
     if ( $urlType eq 'SERVICE' and !( $ticket =~ s/^ST-// ) ) {
         $self->logger->error("Provided ticket is not a service ticket (ST)");
-        return $self->returnCasServiceValidateError( $req, 'INVALID_TICKET',
-            'Provided ticket is not a service ticket' );
+        return $self->_failValidate2(
+            $req,
+            code => 'INVALID_TICKET',
+            msg  => 'Provided ticket is not a service ticket'
+        );
     }
     elsif ( $urlType eq 'PROXY' and !( $ticket =~ s/^(P|S)T-// ) ) {
         $self->userLogger->error(
             "Provided ticket is not a service or proxy ticket ($1T)");
-        return $self->returnCasServiceValidateError( $req, 'INVALID_TICKET',
-            'Provided ticket is not a service or proxy ticket' );
+        return $self->_failValidate2(
+            $req,
+            code => 'INVALID_TICKET',
+            msg  => 'Provided ticket is not a service or proxy ticket'
+        );
     }
 
     my $casServiceSession = $self->getCasSession($ticket);
 
     unless ($casServiceSession) {
         $self->logger->error("$urlType ticket session $ticket not found");
-        return $self->returnCasServiceValidateError( $req, 'INVALID_TICKET',
-            'Ticket not found' );
+        return $self->_failValidate2(
+            $req,
+            code => 'INVALID_TICKET',
+            msg  => 'Ticket not found'
+        );
     }
+
+    # Make sure the token is still valid, we already compensated for
+    # different TTLs when storing _utime
+    if ( $casServiceSession->{data}->{_utime} ) {
+        if (
+            time >
+            ( $casServiceSession->{data}->{_utime} + $self->conf->{timeout} ) )
+        {
+            $self->logger->error("$urlType ticket session $ticket has expired");
+            return $self->_failValidate2(
+                $req,
+                code => 'INVALID_TICKET',
+                msg  => 'Ticket expired'
+            );
+        }
+    }
+
     my $app = $casServiceSession->data->{_casApp};
 
     $self->logger->debug("$urlType ticket session $ticket found");
@@ -700,8 +816,8 @@ sub _validate2 {
         if (   $service1_uri->rel($service2_uri) eq "./"
             or $service2_uri->rel($service1_uri) eq "./" )
         {
-            $self->logger->notice(
-"Submitted service $service1_uri does not exactly match initial service "
+            $self->logger->notice( "Submitted service $service1_uri"
+                  . " does not exactly match initial service "
                   . $service2_uri
                   . ' but difference is tolerated.' );
         }
@@ -710,9 +826,12 @@ sub _validate2 {
                 "Submitted service $service does not match initial service "
                   . $casServiceSession->data->{service} );
             $self->deleteCasSession($casServiceSession);
-            return $self->returnCasServiceValidateError( $req,
-                'INVALID_SERVICE',
-                'Submitted service does not match initial service' );
+            return $self->_failValidate2(
+                $req,
+                app  => $app,
+                code => 'INVALID_SERVICE',
+                msg  => 'Submitted service does not match initial service'
+            );
         }
     }
     else {
@@ -727,11 +846,16 @@ sub _validate2 {
         $self->logger->debug("Renew flag detected ");
 
         unless ( $casServiceSession->data->{renew} ) {
-            $self->logger->error(
-"Authentication renew requested, but not done in former authentication process"
-            );
+            $self->logger->error( "Authentication renew requested,"
+                  . " but not done in former authentication process" );
             $self->deleteCasSession($casServiceSession);
-            return $self->returnCasValidateError();
+            return $self->_failValidate2(
+                $req,
+                app  => $app,
+                code => 'INVALID_TICKET',
+                msg  => 'Authentication renew requested,'
+                  . ' but not done in former authentication process'
+            );
         }
 
     }
@@ -752,7 +876,7 @@ sub _validate2 {
         $PGinfos->{type}    = 'casProxyGranting';
         $PGinfos->{service} = $service;
         $PGinfos->{_cas_id} = $casServiceSession->data->{_cas_id};
-        $PGinfos->{_utime}  = $casServiceSession->data->{_utime};
+        $PGinfos->{_utime}  = time;
         $PGinfos->{_casApp} = $app;
 
         # Trace proxies
@@ -780,9 +904,8 @@ sub _validate2 {
 
                 $casProxyGrantingTicketIOU = "PGTIOU-" . $tmpCasSession->id;
                 $self->deleteCasSession($tmpCasSession);
-                $self->logger->debug(
-"Generate proxy granting ticket IOU $casProxyGrantingTicketIOU"
-                );
+                $self->logger->debug( "Generate proxy granting ticket IOU"
+                      . $casProxyGrantingTicketIOU );
 
                 # Request pgtUrl
                 if (
@@ -818,13 +941,20 @@ sub _validate2 {
               . $casServiceSession->data->{_cas_id}
               . " notfound" );
         $self->deleteCasSession($casServiceSession);
-        return $self->returnCasServiceValidateError( $req, 'INTERNAL_ERROR',
-            'No session associated to ticket' );
+        return $self->_failValidate2(
+            $req,
+            app  => $app,
+            code => 'INTERNAL_ERROR',
+            msg  => 'No session associated to ticket'
+        );
     }
 
     # Get username
     my $username = $self->getUsernameForApp( $req, $app, $localSession->data );
 
+    # Useful for logging
+    $req->sessionInfo( $localSession->data );
+
     $self->logger->debug("Get username $username");
 
     # Get attributes [CAS 3.0]
@@ -858,11 +988,43 @@ sub _validate2 {
     # Return success message
     $self->deleteCasSession($casServiceSession);
 
+    if ( $self->isLogoutEnabled($app) ) {
+        $self->_save_ticket_for_single_logout( $localSession, $service,
+            $req->param('ticket') );
+    }
+
+    $self->p->registerProtectedAppAccess( $req, $username, "cas:$app" );
+
     my $h =
       $self->p->processHook( $req, 'casGenerateValidateResponse', $username,
         $attributes );
-    return $self->returnCasValidateError() if ( $h != PE_OK );
+    return $self->_failValidate2(
+        $req,
+        app  => $app,
+        code => "INTERNAL_ERROR",
+        msg  => "Internal error during request validation",
+        res  => $h,
+    ) if ( $h != PE_OK );
 
+    my $user   = $req->sessionInfo->{ $self->conf->{whatToTrace} };
+    my $by_app = $app ? " by $app" : "";
+    my $attr_str =
+        ( defined($attributes) && keys %$attributes )
+      ? ( " with attributes " . join( ',', sort( keys(%$attributes) ) ) )
+      : "";
+    $self->auditLog(
+        $req,
+        code    => "ISSUER_CAS_VALIDATE_SUCCESS",
+        message => (
+                "CAS validation succeeded for ${user}${by_app}"
+              . " as ${username}${attr_str}"
+        ),
+        ( $app ? ( app => $app ) : () ),
+        user        => $user,
+        cas_user    => $username,
+        cas_proxies => $proxies,
+        attributes  => $attributes,
+    );
     return $self->returnCasServiceValidateSuccess( $req, $username,
         $casProxyGrantingTicketIOU, $proxies, $attributes );
 }
@@ -893,4 +1055,226 @@ sub getUsernameForApp {
     return $username;
 }
 
+sub _save_ticket_for_single_logout {
+    my ( $self, $localSession, $service, $ticket ) = @_;
+
+    $self->logger->debug("Saving service $service ticket $ticket for SLO");
+
+    my $service2ticket = $localSession->data->{cas_tickets_for_SLO} || {};
+    $service2ticket->{$service} = $ticket;
+    $localSession->update( { cas_tickets_for_SLO => $service2ticket } );
+}
+
+sub _send_back_channel_LogoutRequests {
+    my ( $self, $req, $sessionData ) = @_;
+
+    my $service2ticket = $sessionData->{cas_tickets_for_SLO} or return;
+
+    my $content;
+    foreach my $service ( keys %$service2ticket ) {
+
+        # Create a new relay session
+        my $relayInfos = {
+            service => $service,
+            ticket  => $service2ticket->{$service},
+            _type   => "relay"
+        };
+        my $relaySession = $self->getCasSession( undef, $relayInfos );
+        my $relayid      = $relaySession->id;
+
+        # Build the URL that could be used to play this logout request
+        my $slo_url =
+          $self->p->buildUrl( $req->portal, "cas", "relayLogout",
+            { relay => $relayid } );
+
+        my $name = $self->getNameFromService($service);
+
+        # Display information to the user
+        $content .= $self->loadTemplate(
+            $req,
+            'samlSpSoapLogout',
+            params => {
+                imgUrl => $slo_url,
+                name   => $name,
+            }
+        );
+    }
+
+    # Build Relay URL and display it
+    $req->info(
+        $self->loadTemplate(
+            $req, 'samlSpsLogout', params => { content => $content }
+        )
+    ) if $content;
+}
+
+sub getNameFromService {
+    my ( $self, $service ) = @_;
+
+    my $app = $self->getCasApp($service);
+    if ( $self->casAppList->{$app}->{casAppMetaDataOptionsDisplayName} ) {
+        return $self->casAppList->{$app}->{casAppMetaDataOptionsDisplayName};
+    }
+    return $app;
+}
+
+sub relayLogout {
+    my ( $self, $req ) = @_;
+
+    my $relay = $req->param("relay");
+    unless ($relay) {
+        $self->logger->error("Missing relay parameter");
+        return $self->p->imgnok($req);
+    }
+
+    my $session = $self->getCasSession($relay);
+    unless ($session) {
+        $self->logger->error("CAS session $relay not found");
+        return $self->p->imgnok($req);
+    }
+
+    unless ( $session->data->{_type} eq "relay" ) {
+        $self->logger->error( "CAS session $relay has incorrect type "
+              . $session->data->{_type} );
+        return $self->p->imgnok($req);
+    }
+
+    my $service = $session->data->{service};
+    my $ticket  = $session->data->{ticket};
+    $self->logger->debug(
+        "Sending back-channel LogoutRequest to $service (ticket $ticket)");
+
+    my $logoutRequest = $self->buildLogoutRequest($ticket);
+
+    $self->logger->debug("Sending back-channel LogoutRequest to $service");
+    my $response = $self->ua->post(
+        $service,
+        { logoutRequest => $logoutRequest },
+        "Content-Type" => 'application/x-www-form-urlencoded'
+    );
+
+    if ( $response->is_success ) {
+        $self->logger->debug("CAS back-channel logout to $service OK");
+        return $self->p->imgok($req);
+    }
+    else {
+        $self->logger->error( "CAS back-channel logout to $service error: "
+              . $response->message );
+        $self->logger->debug( $response->dump );
+        return $self->p->imgnok($req);
+    }
+}
+
+sub isLogoutEnabled {
+    my ( $self, $app ) = @_;
+
+    my $option = -1;
+    if ($app) {
+        $option = $self->casAppList->{$app}->{casAppMetaDataOptionsLogout}
+          // -1;
+    }
+
+    if ( $option == -1 ) {
+        return $self->conf->{casBackChannelSingleLogout} || 0;
+    }
+    else {
+        return $option;
+    }
+}
+
+sub _failValidate1 {
+    my ( $self, $req, %params ) = @_;
+    my $reason = $params{'msg'} ? ": $params{'msg'}" : "";
+    my $res    = $params{res} || PE_ERROR;
+    my $app    = $params{app};
+
+    $self->auditLog(
+        $req,
+        code    => "ISSUER_CAS_VALIDATE_FAILED",
+        message => ( "CAS validation failed" . $reason ),
+        ( $reason ? ( reason => $reason ) : () ),
+        ( $app    ? ( app    => $app )    : () ),
+        portal_error => portalConsts->{$res},
+    );
+
+    return $self->returnCasValidateError();
+}
+
+sub _failValidate2 {
+    my ( $self, $req, %params ) = @_;
+    my $reason = $params{'msg'} ? ": $params{'msg'}" : "";
+    my $code   = $params{'code'} || 'INTERNAL_ERROR';
+    my $res    = $params{res}    || PE_ERROR;
+    my $app    = $params{app};
+
+    $self->auditLog(
+        $req,
+        code    => "ISSUER_CAS_VALIDATE_FAILED",
+        message => ( "CAS validation failed" . $reason ),
+        ( $reason ? ( reason => $reason ) : () ),
+        ( $app    ? ( app    => $app )    : () ),
+        cas_code     => $code,
+        portal_error => portalConsts->{$res},
+    );
+
+    return $self->returnCasServiceValidateError( $req, $code, $reason );
+}
+
+sub _failLogin {
+    my ( $self, $req, %params ) = @_;
+    my $reason = $params{'msg'} ? ": $params{'msg'}" : "";
+    my $res    = $params{res} || PE_ERROR;
+    my $app    = $params{app};
+
+    $self->auditLog(
+        $req,
+        code    => "ISSUER_CAS_LOGIN_FAILED",
+        message => ( "CAS login failed" . $reason ),
+        ( $reason ? ( reason => $reason ) : () ),
+        ( $app    ? ( app    => $app )    : () ),
+        portal_error => portalConsts->{$res},
+    );
+
+    return $res;
+}
+
+sub _failLoginRule {
+    my ( $self, $req, $cas_request, $app ) = @_;
+    my $casAccessControlPolicy = $self->conf->{casAccessControlPolicy};
+
+    my $user = $req->sessionInfo->{ $self->conf->{whatToTrace} };
+    $self->auditLog(
+        $req,
+        code => "ISSUER_CAS_LOGIN_FAILED",
+        ( $app ? ( app => $app ) : () ),
+        message => ("User $user is not authorized to access to $app"),
+        reason  => "User is not authorized by access rule",
+    );
+
+    if ( $casAccessControlPolicy =~ /^(error)$/i ) {
+        return PE_UNAUTHORIZEDPARTNER;
+    }
+    else {
+        return $self->_redirectUser( $req, $cas_request->{service},
+            "ST-F4K3T1CK3T" );
+    }
+}
+
+sub _redirectUser {
+    my ( $self, $req, $destination, $ticket ) = @_;
+
+    # Redirect to service
+    # cas_request may have been modified by hook
+    if ($ticket) {
+        $destination .= ( $destination =~ /\?/ ? '&' : '?' )
+          . build_urlencoded( ticket => $ticket );
+    }
+
+    $self->logger->debug("Redirect user to $destination");
+
+    $req->{urldc} = $destination;
+    $req->steps( [] );
+    return PE_OK;
+}
+
 1;
