@@ -7,59 +7,61 @@ use Net::LDAP;    #inherits
 use Net::LDAP::Util qw(escape_filter_value);
 use base qw(Net::LDAP);
 use Lemonldap::NG::Portal::Main::Constants ':all';
+use Net::LDAP::Control::PasswordPolicy;
 use Encode;
-use Unicode::String qw(utf8);
 use Scalar::Util 'weaken';
 use IO::Socket::Timeout;
+use Net::LDAP qw(LDAP_PP_PASSWORD_EXPIRED LDAP_PP_ACCOUNT_LOCKED
+                 LDAP_PP_CHANGE_AFTER_RESET LDAP_PP_PASSWORD_MOD_NOT_ALLOWED
+                 LDAP_PP_MUST_SUPPLY_OLD_PASSWORD LDAP_PP_INSUFFICIENT_PASSWORD_QUALITY
+                 LDAP_PP_PASSWORD_TOO_SHORT LDAP_PP_PASSWORD_TOO_YOUNG
+                 LDAP_PP_PASSWORD_IN_HISTORY );
+
 use utf8;
 
-our $VERSION  = '2.0.10';
-our $ppLoaded = 0;
-
-BEGIN {
-    eval {
-        require threads::shared;
-        threads::shared::share($ppLoaded);
-    };
-}
+our $VERSION = '2.17.0';
 
 # INITIALIZATION
 
 # Build a Net::LDAP object using parameters issued from $portal
 sub new {
     my ( $class, $args ) = @_;
-    my $portal = $args->{p}    or die "$class : p argument required !";
-    my $conf   = $args->{conf} or die "$class : conf argument required !";
-    my $self;
-    my $useTls = 0;
-    my $tlsParam;
-    my @servers = ();
-    foreach my $server ( split /[\s,]+/, $conf->{ldapServer} ) {
+    my $portal = $args->{p}    or die "$class: p argument is required!";
+    my $conf   = $args->{conf} or die "$class: conf argument is required!";
+    my ( $self, @servers, %tlsParams );
+    my $useStartTls = 0;
 
+    foreach my $server ( split /[\s,]+/, $conf->{ldapServer} ) {
         if ( $server =~ m{^ldap\+tls://([^/]+)/?\??(.*)$} ) {
-            $useTls   = 1;
-            $server   = $1;
-            $tlsParam = $2 || "";
+            $useStartTls = 1;
+            $server      = $1;
+            %tlsParams   = split( /[&=]/, $2 || "" );
+        }
+        elsif ( $server =~ m{^(ldaps://[^/]+)/?\??(.*)$} ) {
+            $useStartTls = 0;
+            $server      = $1;
+            %tlsParams   = split( /[&=]/, $2 || "" );
         }
         else {
-            $useTls = 0;
+            $useStartTls = 0;
         }
         push @servers, $server;
     }
+    $tlsParams{cafile} ||= $conf->{ldapCAFile} if $conf->{ldapCAFile};
+    $tlsParams{capath} ||= $conf->{ldapCAPath} if $conf->{ldapCAPath};
+    $tlsParams{verify} ||= $conf->{ldapVerify} if $conf->{ldapVerify};
     $self = Net::LDAP->new(
         \@servers,
         onerror   => undef,
         keepalive => 1,
+        %tlsParams,
         ( $conf->{ldapPort}    ? ( port    => $conf->{ldapPort} )    : () ),
         ( $conf->{ldapTimeout} ? ( timeout => $conf->{ldapTimeout} ) : () ),
         ( $conf->{ldapVersion} ? ( version => $conf->{ldapVersion} ) : () ),
         ( $conf->{ldapRaw}     ? ( raw     => $conf->{ldapRaw} )     : () ),
-        ( $conf->{ldapCAFile}  ? ( cafile  => $conf->{ldapCAFile} )  : () ),
-        ( $conf->{ldapCAPath}  ? ( capath  => $conf->{ldapCAPath} )  : () ),
-        ( $conf->{ldapVerify}  ? ( verify  => $conf->{ldapVerify} )  : () ),
     );
     unless ($self) {
-        $portal->logger->error($@);
+        $portal->logger->error( "LDAP initialization error: " . $@ );
         return 0;
     }
     elsif ( $Net::LDAP::VERSION < '0.64' ) {
@@ -72,7 +74,7 @@ sub new {
             and $self->socket->errstr < 0 )
         {
             $portal->logger->error(
-                "SSL connection error: " . $self->socket->errstr );
+                "LDAP SSL connection failed: " . $self->socket->errstr );
             return 0;
         }
     }
@@ -84,14 +86,10 @@ sub new {
     $socket->read_timeout( $conf->{ldapIOTimeout} );
     $socket->write_timeout( $conf->{ldapIOTimeout} );
 
-    if ($useTls) {
-        my %h = split( /[&=]/, $tlsParam );
-        $h{cafile} ||= $conf->{ldapCAFile} if ( $conf->{ldapCAFile} );
-        $h{capath} ||= $conf->{ldapCAPath} if ( $conf->{ldapCAPath} );
-        $h{verify} ||= $conf->{ldapVerify} if ( $conf->{ldapVerify} );
-        my $mesg = $self->start_tls(%h);
+    if ($useStartTls) {
+        my $mesg = $self->start_tls(%tlsParams);
         if ( $mesg->code ) {
-            $portal->logger->error( 'StartTLS failed: ' . $mesg->error );
+            $portal->logger->error( 'LDAP StartTLS failed: ' . $mesg->error );
             return 0;
         }
     }
@@ -115,10 +113,9 @@ sub new {
 # @return Net::LDAP::Message
 sub bind {
     my ( $self, $dn, %args ) = @_;
+    my $mesg;
 
     $self->{portal}->logger->debug("Call bind for $dn") if $dn;
-
-    my $mesg;
     unless ($dn) {
         $dn = $self->{conf}->{managerDn};
         $args{password} =
@@ -135,7 +132,31 @@ sub bind {
             };
             print STDERR "$@\n" if ($@);
         }
+
+        if ( $self->{conf}->{ldapPpolicyControl} ) {
+            my $pp = Net::LDAP::Control::PasswordPolicy->new();
+            $args{control} = [$pp];
+        }
+
         $mesg = $self->SUPER::bind( $dn, %args );
+
+        if ( $mesg->code ) {
+            my ($resp) = $mesg->control("1.3.6.1.4.1.42.2.27.8.5.1");
+
+            # Check for ppolicy error
+            my ( $llng_code, $ppolicy_error_code, $ppolicy_error_msg ) = $self->check_ppolicy($resp);
+            if ( $llng_code != 0 ) {
+                $self->{portal}
+                  ->logger->error( "Error when service account $dn binding to LDAP server: "
+                      . $mesg->error
+                      . " | extended ppolicy control response error: $ppolicy_error_msg"
+                  );
+            }
+            else {
+                $self->{portal}->logger->error(
+                    "Error when binding to LDAP server: " . $mesg->error );
+            }
+        }
     }
     else {
         $mesg = $self->SUPER::bind();
@@ -143,6 +164,39 @@ sub bind {
     return $mesg;
 }
 
+# check_password policy error code thanks to ldap response
+# input: ldap response
+# return: ( llng_code, ppolicy_error_code, ppolicy_error_msg )
+sub check_ppolicy {
+    my $self = shift;
+    my $resp = shift;
+    return (PE_OK, undef, undef) unless(defined($resp));
+
+    my $PPOLICY_ERRORS = {
+        LDAP_PP_PASSWORD_EXPIRED() => [ PE_PP_PASSWORD_EXPIRED, "password expired" ],
+        LDAP_PP_ACCOUNT_LOCKED() => [ PE_PP_ACCOUNT_LOCKED, "account locked" ],
+        LDAP_PP_CHANGE_AFTER_RESET() => [ PE_PP_CHANGE_AFTER_RESET, "change after reset" ],
+        LDAP_PP_PASSWORD_MOD_NOT_ALLOWED() => [ PE_PP_PASSWORD_MOD_NOT_ALLOWED, "password mod not allowed" ],
+        LDAP_PP_MUST_SUPPLY_OLD_PASSWORD() => [ PE_PP_MUST_SUPPLY_OLD_PASSWORD, "supply old password" ],
+        LDAP_PP_INSUFFICIENT_PASSWORD_QUALITY() => [ PE_PP_INSUFFICIENT_PASSWORD_QUALITY, "insufficient password quality" ],
+        LDAP_PP_PASSWORD_TOO_SHORT() => [ PE_PP_PASSWORD_TOO_SHORT, "password too short" ],
+        LDAP_PP_PASSWORD_TOO_YOUNG() => [ PE_PP_PASSWORD_TOO_YOUNG, "password too young" ],
+        LDAP_PP_PASSWORD_IN_HISTORY() => [ PE_PP_PASSWORD_IN_HISTORY, "password in history" ]
+    };
+
+    my $ppolicy_error_code = $resp->pp_error;
+
+    if ( defined $ppolicy_error_code and $PPOLICY_ERRORS->{$ppolicy_error_code} ) {
+
+        # return: ( llng_code, ppolicy_error_code, ppolicy_error_msg )
+        return ( $PPOLICY_ERRORS->{$ppolicy_error_code}->[0],
+                 $ppolicy_error_code,
+                 $PPOLICY_ERRORS->{$ppolicy_error_code}->[1]
+               );
+    }
+    return (PE_OK, undef, undef);
+}
+
 ## @method Net::LDAP::Message unbind()
 # Reimplementation of Net::LDAP::unbind() to force call to disconnect()
 # @return Net::LDAP::Message
@@ -158,30 +212,6 @@ sub unbind {
     return $mesg;
 }
 
-## @method private boolean loadPP ()
-# Load Net::LDAP::Control::PasswordPolicy
-# @return true if succeed.
-sub loadPP {
-    my $self = shift;
-    return 1 if ($ppLoaded);
-
-    # Minimal version of Net::LDAP required
-    if ( $Net::LDAP::VERSION < 0.38 ) {
-        die(
-"Module Net::LDAP is too old for password policy, please install version 0.38 or higher"
-        );
-    }
-
-    # Require Perl module
-    eval { require Net::LDAP::Control::PasswordPolicy };
-    if ($@) {
-        $self->{portal}->logger->error(
-            "Module Net::LDAP::Control::PasswordPolicy not found in @INC");
-        return 0;
-    }
-    $ppLoaded = 1;
-}
-
 ## @method protected int userBind(string dn, hash args)
 # Call bind() with dn/password and return
 # @param $dn LDAP distinguish name
@@ -202,7 +232,7 @@ sub userBind {
         # Get server control response
         my ($resp) = $mesg->control("1.3.6.1.4.1.42.2.27.8.5.1");
 
-        # Return direct unless control resonse
+        # Return direct unless control response
         unless ( defined $resp ) {
             if ( $mesg->code == 49 ) {
                 $self->{portal}->userLogger->warn(
@@ -221,21 +251,11 @@ sub userBind {
         }
 
         # Check for ppolicy error
-        my $pp_error = $resp->pp_error;
-        if ( defined $pp_error ) {
+        my ( $llng_code, $ppolicy_error_code, $ppolicy_error_msg ) = $self->check_ppolicy($resp);
+        if ( $llng_code != 0 ) {
             $self->{portal}->userLogger->error(
-                "Password policy error $pp_error for " . $req->user );
-            return [
-                PE_PP_PASSWORD_EXPIRED,
-                PE_PP_ACCOUNT_LOCKED,
-                PE_PP_CHANGE_AFTER_RESET,
-                PE_PP_PASSWORD_MOD_NOT_ALLOWED,
-                PE_PP_MUST_SUPPLY_OLD_PASSWORD,
-                PE_PP_INSUFFICIENT_PASSWORD_QUALITY,
-                PE_PP_PASSWORD_TOO_SHORT,
-                PE_PP_PASSWORD_TOO_YOUNG,
-                PE_PP_PASSWORD_IN_HISTORY,
-            ]->[$pp_error];
+                "Password policy error $ppolicy_error_code ($ppolicy_error_msg) for " . $req->user );
+            return $llng_code;
         }
         elsif ( $mesg->code == 0 ) {
 
@@ -321,10 +341,11 @@ sub userModifyPassword {
         $passwordAttribute = "unicodePwd";
 
         # Encode password for AD
-        $newpassword = utf8( chr(34) . $newpassword . chr(34) )->utf16le();
+        $newpassword = encode( "UTF-16LE",
+            decode( "UTF-8", chr(34) . $newpassword . chr(34) ) );
         if ( $oldpassword and $asUser ) {
-            $oldpassword =
-              utf8( chr(34) . $oldpassword . chr(34) )->utf16le();
+            $oldpassword = encode( "UTF-16LE",
+                decode( "UTF-8", chr(34) . $oldpassword . chr(34) ) );
         }
         $self->{portal}->logger->debug("Active Directory mode enabled");
 
@@ -623,39 +644,6 @@ sub userModifyPassword {
     }
 }
 
-## @method protected Lemonldap::NG::Portal::_LDAP ldap()
-# @return Lemonldap::NG::Portal::_LDAP object
-sub ldap {
-    my $self = shift;
-    return $self->{ldap}
-      if ( ref( $self->{ldap} )
-        and $self->{flags}->{ldapActive} );
-    if ( $self->{ldap} = Lemonldap::NG::Portal::_LDAP->new($self)
-        and my $mesg = $self->{ldap}->bind )
-    {
-        if ( $mesg->code != 0 ) {
-            $self->logger->error( "LDAP error: " . $mesg->error );
-            $self->{ldap}->unbind;
-        }
-        else {
-            if ( $self->{ldapPpolicyControl}
-                and not $self->{ldap}->loadPP() )
-            {
-                $self->logger->error("LDAP password policy error");
-                $self->{ldap}->unbind;
-            }
-            else {
-                $self->{flags}->{ldapActive} = 1;
-                return $self->{ldap};
-            }
-        }
-    }
-    else {
-        $self->logger->error("LDAP error: $@");
-    }
-    return 0;
-}
-
 ## @method string searchGroups(string base, string key, string value, string attributes, hashref dupcheck)
 # Get groups from LDAP directory
 # @param base LDAP search base
@@ -671,8 +659,9 @@ sub searchGroups {
     my $groups = {};
 
     # Creating search filter
+    my $groupSearchFilter = $self->{conf}->{groupLDAPFilter} || "(objectClass=" . $self->{conf}->{ldapGroupObjectClass} . ")";
     my $searchFilter =
-      "(&(objectClass=" . $self->{conf}->{ldapGroupObjectClass} . ")(|";
+      "(&" . $groupSearchFilter . "(|";
     foreach ( split( $self->{conf}->{multiValuesSeparator}, $value ) ) {
         $searchFilter .= "(" . $key . "=" . escape_filter_value($_) . ")";
     }
