@@ -1,12 +1,13 @@
 package Lemonldap::NG::Portal::Main::Process;
 
-our $VERSION = '2.0.15';
+our $VERSION = '2.19.0';
 
 package Lemonldap::NG::Portal::Main;
 
 use strict;
 use MIME::Base64;
 use POSIX qw(strftime);
+use Lemonldap::NG::Common::Util qw(isHiddenAttr);
 use Lemonldap::NG::Portal::Main::Constants qw(portalConsts);
 use URI;
 
@@ -21,19 +22,33 @@ sub process {
 
     # Store ipAddr in env
     $req->env->{ipAddr} = $req->address;
-    my $err = PE_OK;
+    my $err    = PE_OK;
+    my $nofail = delete $args{nofail};
     while ( my $sub = shift @{ $req->steps } ) {
+        $nofail = 1 if $req->data->{nofail};
         if ( ref $sub ) {
             $self->logger->debug("Processing code ref");
-            last if ( $err = $sub->( $req, %args ) );
+            my $tmp = $sub->( $req, %args );
+            if ($tmp) {
+                $err = $tmp unless $err > 0;
+                last        unless $nofail and $err != PE_SENDRESPONSE;
+            }
         }
         else {
             $self->logger->debug("Processing $sub");
             if ( my $as = $self->aroundSub->{$sub} ) {
-                last if ( $err = $as->( $req, %args ) );
+                my $tmp = $as->( $req, %args );
+                if ($tmp) {
+                    $err = $tmp unless $err > 0;
+                    last        unless $nofail and $err != PE_SENDRESPONSE;
+                }
             }
             else {
-                last if ( $err = $self->$sub( $req, %args ) );
+                my $tmp = $self->$sub( $req, %args );
+                if ($tmp) {
+                    $err = $tmp unless $err > 0;
+                    last        unless $nofail and $err != PE_SENDRESPONSE;
+                }
             }
             if ( $self->afterSub->{$sub} ) {
                 unshift @{ $req->steps }, @{ $self->afterSub->{$sub} };
@@ -61,7 +76,7 @@ sub processHook {
         }
     }
     if ( $err != PE_OK ) {
-        my $msg = "Hook $hookName returned " . portalConsts->{$err};
+        my $msg = "Hook $hookName returned $err (" . portalConsts->{$err} . ")";
         if ( $err > 0 ) {
             $self->logger->warn($msg);
         }
@@ -132,7 +147,7 @@ sub controlUrl {
         else {
             if ( $url =~ m#[^A-Za-z0-9\+/=]# ) {
                 unless ( $req->maybeNotBase64 ) {
-                    $self->userLogger->error(
+                    $self->logger->error(
 "Value must be BASE64 encoded (param: url | value: $url)"
                     );
                     return PE_BADURL;
@@ -144,29 +159,24 @@ sub controlUrl {
                 $req->{urldc} =~ s/[\r\n]//sg;
             }
         }
+        $req->{urldc} = URI->new( $req->{urldc} )->as_string;
 
-        # For logout request, test if Referer comes from an authorized site
-        my $tmp = (
-              $req->param('logout')
-            ? $req->referer
-            : $req->{urldc}
-        );
+        my $tmp = $req->urldc;
 
         # XSS attack
-        if (
-            $self->checkXSSAttack(
-                $req->param('logout') ? 'HTTP Referer' : 'urldc',
-                $req->{urldc}
-            )
-          )
-        {
+        if ( $self->checkXSSAttack( 'urldc', $req->{urldc} ) ) {
             delete $req->{urldc};
             return PE_BADURL;
         }
 
         # Unprotected hosts
         if ( $tmp and ( $tmp !~ URIRE ) ) {
-            $self->userLogger->error("Bad URL $tmp");
+            $self->auditLog(
+                $req,
+                message => "Bad URL $tmp",
+                code    => "UNAUTHORIZED_REDIRECT",
+                url     => $tmp,
+            );
             delete $req->{urldc};
             return PE_BADURL;
         }
@@ -183,22 +193,31 @@ sub controlUrl {
         # Try to resolve alias
         my $originalVhost = $self->HANDLER->resolveAlias($vhost);
         $vhost = "$proto://$originalVhost";
-        $self->logger->debug( 'Required URL (param: '
-              . ( $req->param('logout') ? 'HTTP Referer'   : 'urldc' )
+        $self->logger->debug( 'Required URL (param: urldc'
               . ( $tmp ? " | value: $tmp | alias: $vhost)" : ')' ) );
 
         # If the target URL has an authLevel set in config, remember it.
         my $level = $self->HANDLER->getLevel( $req, $appuri, $originalVhost );
-        $req->pdata->{targetAuthnLevel} = $level if $level;
+
+        if ($level) {
+            $self->logger->debug( "VirtualHost $originalVhost requires"
+                  . " authentication level $level" );
+            $req->pdata->{targetAuthnLevel} = $level;
+        }
 
         if (    $tmp
             and !$self->isTrustedUrl($tmp)
             and !$self->isTrustedUrl($vhost) )
         {
-            $self->userLogger->error(
-                    "URL contains an unprotected host (param: "
-                  . ( $req->param('logout') ? 'HTTP Referer' : 'urldc' )
-                  . " | value: $tmp | alias: $vhost)" );
+            $self->auditLog(
+                $req,
+                message => (
+                        "URL contains an unprotected host (param: urldc"
+                      . " | value: $tmp | alias: $vhost)"
+                ),
+                code => "UNAUTHORIZED_REDIRECT",
+                url  => $tmp,
+            );
             delete $req->{urldc};
             return PE_UNPROTECTEDURL;
         }
@@ -214,6 +233,7 @@ sub controlUrl {
 sub checkLogout {
     my ( $self, $req ) = @_;
     if ( defined $req->param('logout') ) {
+        $req->data->{nofail} = 1;
         $req->steps(
             [ @{ $self->beforeLogout }, 'authLogout', 'deleteSession' ] );
     }
@@ -223,20 +243,19 @@ sub checkLogout {
 sub checkUnauthLogout {
     my ( $self, $req ) = @_;
     if ( defined $req->param('logout') ) {
-        $self->userLogger->info('Unauthenticated logout request');
-        $self->logger->debug('Cleaning pdata');
-        $self->logger->debug("Removing $self->{conf}->{cookieName} cookie");
-        $req->pdata( {} );
-        $req->addCookie(
-            $self->cookie(
-                name    => $self->conf->{cookieName},
-                domain  => $self->conf->{domain},
-                secure  => $self->conf->{securedCookie},
-                expires => 'Wed, 21 Oct 2015 00:00:00 GMT',
-                value   => 0
-            )
-        );
-        $req->steps( [ sub { PE_LOGOUT_OK } ] );
+        $self->_unauthLogout($req);
+        $req->steps( [ 'controlUrl', sub { PE_LOGOUT_OK } ] );
+    }
+    return PE_OK;
+}
+
+sub checkCancel {
+    my ( $self, $req ) = @_;
+    if ( $req->param('cancel') ) {
+        $self->logger->debug('Cancel called, run authCancel calls');
+        $req->mustRedirect(1);
+        $req->steps( [ @{ $self->authCancel }, sub { PE_OK } ] );
+        return PE_OK;
     }
     return PE_OK;
 }
@@ -314,7 +333,7 @@ sub deleteSession {
         # Redirect on logout page if no other target defined
         if ( !$req->urldc and !$req->postUrl ) {
             $self->logger->debug('No other target defined, redirect on logout');
-            $req->urldc( $self->buildUrl( { logout => 1 } ) );
+            $req->urldc( $self->buildUrl( $req->portal, { logout => 1 } ) );
         }
     }
     $req->userData( {} );
@@ -325,14 +344,14 @@ sub deleteSession {
         return PE_OK;
     }
 
-    if ( $req->urldc and $req->urldc ne $self->conf->{portal} ) {
+    if ( $req->urldc and $req->urldc ne $req->portal ) {
         $req->steps( [] );
         return PE_REDIRECT;
     }
 
     # If logout redirects to another URL, just remove next steps for the
     # request so autoRedirect will be called
-    if ( $req->{urldc} and $req->{urldc} ne $self->conf->{portal} ) {
+    if ( $req->{urldc} and $req->{urldc} ne $req->portal ) {
         $req->steps( [] );
         return PE_OK;
     }
@@ -354,6 +373,23 @@ sub checkXSSAttack {
     # Test value
     $value =~ s/\%25/\%/g;
     if ( $value =~ m/(?:\0|<|'|"|`|\%(?:00|3C|22|27))/ ) {
+        $self->logger->error(
+            "XSS attack detected (param: $name | value: $value)");
+        return $self->conf->{checkXSS};
+    }
+    return 0;
+}
+
+# Relaxed version that doesn't look at URL escaped character
+# This is meant to be used for checking urldc until we find a better solution
+sub checkXSSAttackUrldc {
+    my ( $self, $name, $value ) = @_;
+
+    # Empty values are not bad
+    return 0 unless $value;
+
+    # Test value
+    if ( $value =~ m/(?:\0|<|'|"|`)/ ) {
         $self->userLogger->error(
             "XSS attack detected (param: $name | value: $value)");
         return $self->conf->{checkXSS};
@@ -377,10 +413,10 @@ sub extractFormInfo {
         and $req->cookies->{ $self->conf->{cookieName} } )
     {
         $req->addCookie(
-            $self->cookie(
+            $self->genDomainCookie(
+                $req,
                 name    => $self->conf->{cookieName},
                 value   => 0,
-                domain  => $self->conf->{domain},
                 secure  => $self->conf->{securedCookie},
                 expires => 'Wed, 21 Oct 2015 00:00:00 GMT'
             )
@@ -468,7 +504,7 @@ sub setSessionInfo {
     }
 
     # Currently selected language
-    $req->{sessionInfo}->{_language} = $req->cookies->{llnglanguage} || 'en';
+    $req->{sessionInfo}->{_language} = $self->getLanguage($req);
 
     # Store URL origin in session
     $req->{sessionInfo}->{_url} = $req->{urldc};
@@ -549,20 +585,17 @@ sub store {
         $infos{_httpSessionType} = 1;
 
         my $session2 = $self->getApacheSession( undef, info => \%infos );
-        $self->logger->debug("Create second session for unsecured cookie...");
+        $self->logger->debug("Create second session for unsecured cookie");
         $req->{sessionInfo}->{_httpSession} = $session2->id;
-        $self->logger->debug( " -> Cookie value : " . $session2->id );
+        $self->logger->debug( " -> Cookie value: " . $session2->id );
     }
 
     # Fill session
     my $infos = {};
-    foreach my $k ( keys %{ $req->{sessionInfo} } ) {
+    foreach my $k ( sort keys %{ $req->{sessionInfo} } ) {
         next unless defined $req->{sessionInfo}->{$k};
         my $displayValue = $req->{sessionInfo}->{$k};
-        $displayValue = '****'
-          if (  $self->conf->{hiddenAttributes}
-            and $self->conf->{hiddenAttributes} =~ /\b$k\b/ );
-
+        $displayValue = '****' if isHiddenAttr( $self->conf, $k );
         $self->logger->debug("Store $displayValue in session key $k");
         $self->_dump($displayValue) if ref($displayValue);
         $infos->{$k} = $req->{sessionInfo}->{$k};
@@ -571,8 +604,9 @@ sub store {
     # Main session
     my $session = $self->getApacheSession(
         $req->id,
-        force => $req->{force},
-        info  => $infos
+        hashStore => $req->data->{hashStore},
+        force     => $req->{force},
+        info      => $infos
     );
     return PE_APACHESESSIONERROR unless $session;
 
@@ -600,19 +634,19 @@ sub buildCookie {
     my ( $self, $req ) = @_;
     if ( $req->id ) {
         $req->addCookie(
-            $self->cookie(
+            $self->genDomainCookie(
+                $req,
                 name   => $self->conf->{cookieName},
                 value  => $req->id,
-                domain => $self->conf->{domain},
                 secure => $self->conf->{securedCookie},
             )
         );
         if ( $self->conf->{securedCookie} >= 2 ) {
             $req->addCookie(
-                $self->cookie(
+                $self->genDomainCookie(
+                    $req,
                     name   => $self->conf->{cookieName} . "http",
                     value  => $req->{sessionInfo}->{_httpSession},
-                    domain => $self->conf->{domain},
                     secure => 0,
                 )
             );
@@ -623,9 +657,16 @@ sub buildCookie {
         ? $req->{sessionInfo}
         : $req->{userData}
     );
-    $self->userLogger->notice( 'User '
-          . $ref->{ $self->conf->{whatToTrace} }
-          . " successfully authenticated at level $ref->{authenticationLevel}"
+    $self->auditLog(
+        $req,
+        message => (
+                'User '
+              . $ref->{ $self->conf->{whatToTrace} }
+              . " successfully authenticated at level $ref->{authenticationLevel}"
+        ),
+        code  => "AUTHENTICATED",
+        user  => $ref->{ $self->conf->{whatToTrace} },
+        level => $ref->{authenticationLevel},
     );
     return PE_OK;
 }
@@ -643,4 +684,28 @@ sub storeHistory {
     return PE_OK;
 }
 
+# This method returns the domain for a new cookie
+# In modern browsers:
+# - an empty value means the cookie is only valid on the portal itself
+# - in other situations, the cookie is sent to all subdomains
+#   and a leading dot is added for compatibility
+sub getCookieDomain {
+    my ( $self, $req ) = @_;
+
+    if ( my $domain = $self->conf->{domain} ) {
+        if ( $domain eq '#PORTALDOMAIN#' ) {
+            my $host = URI->new( $req->portal )->host;
+            return $host =~ s/^([^\.]*)//r;
+        }
+        elsif ( $domain eq '#PORTAL#' ) {
+            my $host = URI->new( $req->portal )->host;
+            return ".$host";
+        }
+        else {
+            return $domain =~ s/^(?!\.)/./r;
+        }
+    }
+    return;
+}
+
 1;
