@@ -3,14 +3,15 @@ package Lemonldap::NG::Portal::Main::SecondFactor;
 use strict;
 use Mouse;
 use Lemonldap::NG::Portal::Main::Constants qw(
-  PE_SENDRESPONSE
   PE_OK
+  PE_ERROR
   PE_NOTOKEN
+  PE_SENDRESPONSE
   PE_TOKENEXPIRED
   PE_BADCREDENTIALS
 );
 
-our $VERSION = '2.0.6';
+our $VERSION = '2.20.0';
 
 extends qw(
   Lemonldap::NG::Portal::Main::Plugin
@@ -25,7 +26,8 @@ has ott => (
     default => sub {
         my $ott =
           $_[0]->{p}->loadModule('Lemonldap::NG::Portal::Lib::OneTimeToken');
-        $ott->timeout( $_[0]->{conf}->{formTimeout} );
+        $ott->timeout( $_[0]->{conf}->{sfLoginTimeout}
+              || $_[0]->{conf}->{formTimeout} );
         return $ott;
     }
 );
@@ -34,111 +36,51 @@ has prefix  => ( is => 'rw' );
 has logo    => ( is => 'rw', default => '2f.png' );
 has label   => ( is => 'rw' );
 has noRoute => ( is => 'ro' );
+has authnLevel => (
+    is      => 'rw',
+    lazy    => 1,
+    default => sub {
+        return $_[0]->conf->{ $_[0]->prefix . '2fAuthnLevel' };
+    }
+);
+
+has rule => (
+    is      => 'rw',
+    lazy    => 1,
+    default => sub {
+        return $_[0]->conf->{ $_[0]->prefix . '2fActivation' };
+    }
+);
+
+# 'type' field of stored _2fDevices
+# Defaults to the last component of the package name
+# But can be overriden by sfExtra
+has type => (
+    is      => 'rw',
+    default => sub {
+        ( split( '::', ref( $_[0] ) ) )[-1];
+    }
+);
 
 sub init {
     my ($self) = @_;
 
-    # Set logo if overriden
-    $self->logo( $self->conf->{ $self->prefix . "2fLogo" } )
-      if ( $self->conf->{ $self->prefix . "2fLogo" } );
+    # Set logo if overridden
+    $self->logo( $self->conf->{ $self->prefix . '2fLogo' } )
+      if $self->conf->{ $self->prefix . '2fLogo' };
 
     # Set label if provided, translation files will be used otherwise
-    $self->label( $self->conf->{ $self->prefix . "2fLabel" } )
-      if ( $self->conf->{ $self->prefix . "2fLabel" } );
+    $self->label( $self->conf->{ $self->prefix . '2fLabel' } )
+      if $self->conf->{ $self->prefix . '2fLabel' };
 
-    unless ( $self->noRoute ) {
-        $self->logger->debug( 'Adding ' . $self->prefix . '2fcheck routes' );
-        $self->addAuthRoute(
-            $self->prefix . '2fcheck' => '_verify',
-            ['POST']
-        );
-        $self->addAuthRoute(
-            $self->prefix . '2fcheck' => '_redirect',
-            ['GET']
-        );
-        $self->addUnauthRoute(
-            $self->prefix . '2fcheck' => '_verify',
-            ['POST']
-        );
-        $self->addUnauthRoute(
-            $self->prefix . '2fcheck' => '_redirect',
-            ['GET']
-        );
-    }
-    1;
+    return 1;
 }
 
-sub _redirect {
-    my ( $self, $req ) = @_;
-    my $arg = $req->env->{QUERY_STRING};
-    return [
-        302, [ Location => $self->conf->{portal} . ( $arg ? "?$arg" : '' ) ], []
-    ];
-}
-
-sub _verify {
+sub get2fTplParams {
     my ( $self, $req ) = @_;
 
-    my $checkLogins = $req->param('checkLogins');
-    $self->logger->debug("checkLogins set") if ($checkLogins);
-
-    # Check token
-    my $token;
-    unless ( $token = $req->param('token') ) {
-        $self->userLogger->error( $self->prefix . ' 2F access without token' );
-        eval { $self->setSecurity($req) };
-        $req->mustRedirect(1);
-        return $self->p->do( $req, [ sub { PE_NOTOKEN } ] );
-    }
-
-    my $session;
-    unless ( $session = $self->ott->getToken($token) ) {
-        $self->userLogger->info('Token expired');
-        $self->setSecurity($req);
-        return $self->p->do( $req, [ sub { PE_TOKENEXPIRED } ] );
-    }
-
-    # Launch second factor verification
-    my $res = $self->verify( $req, $session );
-
-    # Case error
-    if ($res) {
-        $req->noLoginDisplay(1);
-        $req->sessionInfo($session);
-        $req->id( delete $req->sessionInfo->{_2fRealSession} );
-        $req->urldc( delete $req->sessionInfo->{_2fUrldc} );
-        $req->{sessionInfo}->{_utime} = delete $req->{sessionInfo}->{_2fUtime};
-        $req->authResult(PE_BADCREDENTIALS);
-        return $self->p->do( $req,
-            [ sub { $self->p->storeHistory(@_) }, sub { $res } ] );
-    }
-
-    # Else restore session
-    $req->sessionInfo($session);
-    $req->id( delete $req->sessionInfo->{_2fRealSession} );
-    $req->urldc( delete $req->sessionInfo->{_2fUrldc} );
-    $req->{sessionInfo}->{_utime} = delete $req->{sessionInfo}->{_2fUtime};
-    $req->mustRedirect(1);
-    $self->userLogger->notice( $self->prefix
-          . '2F verification for '
-          . $req->sessionInfo->{ $self->conf->{whatToTrace} } );
-
-    if ( my $l = $self->conf->{ $self->prefix . '2fAuthnLevel' } ) {
-        $self->p->updateSession( $req, { authenticationLevel => $l } );
-    }
-    $req->authResult(PE_SENDRESPONSE);
-    return $self->p->do(
-        $req,
-        [
-            @{ $self->p->afterData },
-            $self->p->validSession,
-            'rebuildCookies',
-            @{ $self->p->endAuth },
-            sub { PE_OK }
-        ]
-    );
+    return $self->p->_sfEngine->get2fTplParams( $req, $self );
 }
-
 1;
 __END__
 
@@ -180,8 +122,8 @@ L<Lemonldap::NG::Portal> second factor plugins.
       # replace the rule to detect if user has registered a device key.
       # The rule must be like this :
       # By example :
-      $self->conf->{u2fActivation} = '$_2fDevices =~ /"type":\s*"U2F"/s'
-      # Optionally, the rule can be : '$_2fDevices and $_2fDevices =~ /"type":\s*"U2F"/s'
+      $self->conf->{totp2fActivation} = '$_2fDevices =~ /"type":\s*"TOTP"/s'
+      # Optionally, the rule can be : '$_2fDevices and $_2fDevices =~ /"type":\s*"TOTP"/s'
       # to avoid warning due to undef variable
       #
       # Required call:
@@ -197,7 +139,7 @@ L<Lemonldap::NG::Portal> second factor plugins.
       # "token"
       ...
       # A LLNG constant must be returned. Example:
-      $req->response($my_psgi_response)
+      $req->response($my_psgi_response);
       return PE_SENDRESPONSE;
   }
   # Required 2nd factor verify method
@@ -205,13 +147,14 @@ L<Lemonldap::NG::Portal> second factor plugins.
       my ( $self, $req, $session ) = @_;
       # Use $req->param('field') to get POST responses
       ...
-      if($result eq $goodResult) {
+      if ($result eq $goodResult) {
         return PE_OK;
       }
       else {
-        return PE_BADCREDENTIALS
+        return PE_BADCREDENTIALS;
       }
   }
+  1;
 
 Enable your plugin in lemonldap-ng.ini, section [portal]:
 
@@ -278,7 +221,7 @@ L<https://gitlab.ow2.org/lemonldap-ng/lemonldap-ng/issues>
 =head1 DOWNLOAD
 
 Lemonldap::NG is available at
-L<http://forge.objectweb.org/project/showfiles.php?group_id=274>
+L<https://lemonldap-ng.org/download>
 
 =head1 COPYRIGHT AND LICENSE
 
