@@ -7,24 +7,22 @@ use Lemonldap::NG::Portal::Main::Constants qw(
   PE_WAIT
 );
 
-our $VERSION = '2.0.10';
+our $VERSION = '2.19.0';
 
 extends 'Lemonldap::NG::Portal::Main::Plugin';
 
 # INITIALIZATION
-use constant aroundSub => { authenticate => 'check' };
+use constant afterSub => { setPersistentSessionInfo => 'run' };
 
 has lockTimes => (
     is      => 'rw',
     isa     => 'ArrayRef',
     default => sub { [] }
 );
-
 has maxAge => (
     is  => 'rw',
     isa => 'Int'
 );
-
 has maxFailed => (
     is  => 'rw',
     isa => 'Int'
@@ -56,14 +54,17 @@ sub init {
         return 0;
     }
 
+    my $maxAge = $self->conf->{bruteForceProtectionMaxAge} || 300;
     if ( $self->conf->{bruteForceProtectionIncrementalTempo} ) {
         my $lockTimes = @{ $self->lockTimes } =
           sort { $a <=> $b }
           map {
-            $_ =~ s/\D//;
-            abs $_ < $self->conf->{bruteForceProtectionMaxLockTime} ? abs $_ : ()
+            my $time = s/\D//r;
+            abs $time < $self->conf->{bruteForceProtectionMaxLockTime}
+              ? abs $time
+              : ()
           }
-          grep { /\d+/ }
+          grep /\d+/,
           split /\s*,\s*/, $self->conf->{bruteForceProtectionLockTimes};
 
         unless ($lockTimes) {
@@ -85,27 +86,22 @@ sub init {
             $lockTimes = $self->conf->{failedLoginNumber};
         }
 
-        my $sum = $self->conf->{bruteForceProtectionMaxAge} *
-          ( 1 + $self->conf->{failedLoginNumber} - $lockTimes );
+        my $sum =
+          $maxAge * ( 1 + $self->conf->{failedLoginNumber} - $lockTimes );
         $sum += $_ foreach @{ $self->lockTimes };
         $self->maxAge($sum);
     }
     else {
-        $self->maxAge( $self->conf->{bruteForceProtectionMaxAge} *
-              ( 1 + $self->maxFailed ) );
+        $self->maxAge( $maxAge * ( 1 + $self->maxFailed ) );
     }
 
     return 1;
 }
 
 # RUNNING METHOD
-sub check {
-    my ( $self, $sub, $req ) = @_;
-    my $now = time;
-    $self->p->setSessionInfo($req);
-    $self->logger->debug("Retrieve $req->{user} logins history");
-    $self->p->setPersistentSessionInfo( $req, $req->{user} );
-
+sub run {
+    my ( $self, $req ) = @_;
+    my $now         = time;
     my $countFailed = my @failedLogins =
       map { ( $now - $_->{_utime} ) <= $self->maxAge ? $_ : () }
       @{ $req->sessionInfo->{_loginHistory}->{failedLogin} };
@@ -115,7 +111,7 @@ sub check {
     my $lastFailedLoginEpoch = $failedLogins[0]->{_utime} || undef;
 
     if ( $self->conf->{bruteForceProtectionIncrementalTempo} ) {
-        return $sub->($req) unless $lastFailedLoginEpoch;
+        return PE_OK unless $lastFailedLoginEpoch;
 
         # Delta between current attempt and last failed login
         my $delta = $now - $lastFailedLoginEpoch;
@@ -148,10 +144,10 @@ sub check {
             $req->lockTime( $waitingTime - $delta );
             return PE_WAIT;
         }
-        return $sub->($req);
+        return PE_OK;
     }
 
-    return $sub->($req)
+    return PE_OK
       if ( $countFailed < $self->maxFailed );
 
     # Delta between current attempt and last failed login
@@ -159,7 +155,7 @@ sub check {
     $self->logger->debug(" -> Delta = $delta");
 
     # Delta < Tempo => wait
-    return $sub->($req)
+    return PE_OK
       unless ( $delta < $self->conf->{bruteForceProtectionTempo}
         && $countFailed );
 
