@@ -4,6 +4,7 @@ use strict;
 use Mouse;
 use JSON qw(from_json to_json);
 use Time::Local;
+use Lemonldap::NG::Common::Session 'id2storage';
 use Lemonldap::NG::Portal::Main::Constants qw(
   PE_OK
   PE_ERROR
@@ -12,7 +13,7 @@ use Lemonldap::NG::Portal::Main::Constants qw(
   PE_SENDRESPONSE
 );
 
-our $VERSION = '2.0.10';
+our $VERSION = '2.19.0';
 
 extends qw(
   Lemonldap::NG::Portal::Main::Plugin
@@ -24,7 +25,7 @@ use constant beforeLogout => 'run';
 
 # INITIALIZATION
 has rule => ( is => 'rw', default => sub { 0 } );
-has ott  => (
+has ott => (
     is      => 'rw',
     lazy    => 1,
     default => sub {
@@ -67,8 +68,14 @@ sub run {
     return PE_OK unless ( $nbr > 1 );
 
     # Force GlobalLogout if timer is disabled
-    unless ( $self->conf->{globalLogoutTimer} ) {
-        $self->logger->debug("GlobalLogout: timer disabled");
+    if ( $req->param('confirm') && $req->param('confirm') == 1
+        || !$self->conf->{globalLogoutTimer} )
+    {
+        my $msg =
+          $self->conf->{globalLogoutTimer}
+          ? 'GlobalLogout: confirm parameter detected'
+          : 'GlobalLogout: timer disabled';
+        $self->logger->debug($msg);
         $self->userLogger->info("GlobalLogout: force global logout for $user");
         $nbr = $self->removeOtherActiveSessions( $req, $sessions );
         $self->userLogger->info("$nbr remaining session(s) removed");
@@ -89,17 +96,19 @@ sub run {
         $req,
         'globallogout',
         params => {
-            PORTAL    => $self->conf->{portal},
-            MAIN_LOGO => $self->conf->{portalMainLogo},
-            SKIN      => $self->p->getSkin($req),
-            LANGS     => $self->conf->{showLanguages},
             SESSIONS  => $sessions,
             TOKEN     => $token,
             LOGIN     => $user,
             CUSTOMPRM => $self->conf->{globalLogoutCustomParam}
         }
     );
-    $req->response($tmp);
+
+    $req->response(
+        $req->wantJSON
+        ? $self->sendJSONresponse( $req,
+            { globalLogout => $nbr, confirmationRequired => 1 } )
+        : $tmp
+    );
 
     return PE_SENDRESPONSE;
 }
@@ -112,6 +121,10 @@ sub globalLogout {
     if ( $req->param('all') ) {
         if ( my $token = $req->param('token') ) {
             if ( $token = $self->ott->getToken($token) ) {
+                my $storeId =
+                  $self->conf->{hashedSessionStore}
+                  ? id2storage( $req->{userData}->{_session_id} )
+                  : $req->{userData}->{_session_id};
 
                 # Read active sessions from token
                 my $sessions = eval { from_json( $token->{sessions} ) };
@@ -126,13 +139,21 @@ sub globalLogout {
                   $req->{userData}->{ $self->{conf}->{whatToTrace} };
                 if ( $req_user eq $user ) {
                     foreach (@$sessions) {
-                        unless ( $as = $self->p->getApacheSession( $_->{id} ) )
+                        unless (
+
+                            # searchOn() returns sessions indexed by their
+                            # storage ID, then it is required to set hashStore
+                            # to 0
+                            $as = $self->p->getApacheSession(
+                                $_->{id}, hashStore => 0,
+                            )
+                          )
                         {
                             $self->userLogger->info(
                                 "GlobalLogout: session $_->{id} expired");
                             next;
                         }
-                        unless ( $req->{userData}->{_session_id} eq $_->{id} ) {
+                        unless ( $storeId eq $_->{id} ) {
                             $self->userLogger->info(
                                 "Remove \"$user\" session: $_->{id}");
                             $as->remove;
@@ -172,49 +193,61 @@ sub activeSessions {
     my $customParam    = $self->conf->{globalLogoutCustomParam} || '';
 
     # Try to retrieve sessions from sessions DB
-    $self->logger->debug('Try to retrieve sessions from DB');
-    my $moduleOptions = $self->conf->{globalStorageOptions} || {};
-    $moduleOptions->{backend} = $self->conf->{globalStorage};
-    $self->logger->debug("Looking for \"$user\" sessions...");
-    $sessions =
-      $self->module->searchOn( $moduleOptions, $self->conf->{whatToTrace},
-        $user );
+    if ($user) {
+        $self->logger->debug('Try to retrieve sessions from DB');
+        my $moduleOptions = $self->conf->{globalStorageOptions} || {};
+        $moduleOptions->{backend} = $self->conf->{globalStorage};
+        $self->logger->debug("Looking for \"$user\" sessions...");
+        $sessions =
+          $self->module->searchOn( $moduleOptions, $self->conf->{whatToTrace},
+            $user );
 
-    $self->logger->debug('Remove non-SSO session(s)...');
-    my $other = 0;
-    foreach ( keys %$sessions ) {
-        unless ( $sessions->{$_}->{_session_kind} eq 'SSO' ) {
-            delete $sessions->{$_};
-            $other++;
+        $self->logger->debug('Skip non-SSO session(s)...');
+        my $other = 0;
+        foreach ( keys %$sessions ) {
+            unless ( $sessions->{$_}->{_session_kind} eq 'SSO' ) {
+                delete $sessions->{$_};
+                $other++;
+            }
         }
+        $self->logger->info("$other non-SSO session(s) skipped")
+          if $other;
+
+        $self->logger->debug('Build an array ref with sessions info...');
+        @$activeSessions =
+          map {
+            my $epoch;
+            my $regex = '^(\d{4})(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})$';
+            if ( $_->{startTime} =~ /$regex/ ) {
+                $epoch = timelocal( $6, $5, $4, $3, $2 - 1, $1 );
+                $_->{startTime} = $epoch;
+            }
+            else {
+                delete $_->{startTime};
+            }
+            if ( $_->{updateTime} ) {
+                if ( $_->{updateTime} =~ /$regex/ ) {
+
+                    $epoch = timelocal( $6, $5, $4, $3, $2 - 1, $1 );
+                    $_->{updateTime} = $epoch;
+                }
+                else {
+                    delete $_->{updateTime};
+                }
+            }
+            $_;
+          }
+          sort { $b->{startTime} cmp $a->{startTime} } map {
+            {
+                id          => $_,
+                customParam => $sessions->{$_}->{$customParam},
+                ipAddr      => $sessions->{$_}->{ipAddr},
+                authLevel   => $sessions->{$_}->{authenticationLevel},
+                startTime   => $sessions->{$_}->{_startTime},
+                updateTime  => $sessions->{$_}->{_updateTime}
+            };
+          } keys %$sessions;
     }
-    $self->logger->info("$other non-SSO session(s) removed")
-      if $other;
-
-    $self->logger->debug('Build an array ref with sessions info...');
-    @$activeSessions =
-      map {
-        my $epoch;
-        my $regex = '^(\d{4})(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})$';
-        $_->{startTime} =~ /$regex/;
-        $epoch = timelocal( $6, $5, $4, $3, $2 - 1, $1 );
-        $_->{startTime} = $epoch;
-        if ( $_->{updateTime} ) {
-            $_->{updateTime} =~ /$regex/;
-            $epoch = timelocal( $6, $5, $4, $3, $2 - 1, $1 );
-            $_->{updateTime} = $epoch;
-        }
-        $_;
-      }
-      sort { $b->{startTime} cmp $a->{startTime} } map { {
-            id          => $_,
-            customParam => $sessions->{$_}->{$customParam},
-            ipAddr      => $sessions->{$_}->{ipAddr},
-            authLevel   => $sessions->{$_}->{authenticationLevel},
-            startTime   => $sessions->{$_}->{_startTime},
-            updateTime  => $sessions->{$_}->{_updateTime}
-        };
-      } keys %$sessions;
 
     return $activeSessions;
 }
@@ -224,12 +257,20 @@ sub removeOtherActiveSessions {
     my $count = 0;
     my $as;
 
+    my $storeId =
+      $self->conf->{hashedSessionStore}
+      ? id2storage( $req->{userData}->{_session_id} )
+      : $req->{userData}->{_session_id};
     foreach (@$sessions) {
-        unless ( $as = $self->p->getApacheSession( $_->{id} ) ) {
+
+        # searchOn() returns sessions indexed by their storage ID, then
+        # it is required to set hashStore to 0
+        unless ( $as = $self->p->getApacheSession( $_->{id}, hashStore => 0, ) )
+        {
             $self->userLogger->info("GlobalLogout: session $_->{id} expired");
             next;
         }
-        unless ( $req->{userData}->{_session_id} eq $_->{id} ) {
+        unless ( $storeId eq $_->{id} ) {
             $self->userLogger->info(
 "Remove \"$req->{userData}->{ $self->conf->{whatToTrace} }\" session: $_->{id}"
             );
