@@ -20,7 +20,7 @@ use Lemonldap::NG::Portal::Main::Constants qw(
   PE_NOTIFICATION
 );
 
-our $VERSION = '2.0.10';
+our $VERSION = '2.19.0';
 
 extends 'Lemonldap::NG::Portal::Main::Plugin';
 
@@ -70,7 +70,7 @@ sub init {
     # Search for configuration options
     my $type = $self->conf->{notificationStorage};
     unless ($type) {
-        $self->error('notificationStorage is not defined, aborting');
+        $self->logger->error('notificationStorage is not defined, aborting');
         return 0;
     }
 
@@ -78,7 +78,7 @@ sub init {
     $type = "Lemonldap::NG::Common::Notifications::$type";
     eval "require $type";
     if ($@) {
-        $self->error(
+        $self->logger->error(
             "Unable to load Lemonldap::NG::Common::Notifications::$type: $@");
         return 0;
     }
@@ -100,7 +100,7 @@ sub init {
           or return 0;
     }
     unless ( eval { $self->module->notifObject( $type->new($prms) ); } ) {
-        $self->error($@);
+        $self->logger->error($@);
         return 0;
     }
 
@@ -141,7 +141,7 @@ sub checkNotifDuringAuth {
             and $req->env->{PATH_INFO} ne '/' )
         {
             $req->data->{_url} =
-              encode_base64( $self->conf->{portal} . $req->env->{PATH_INFO},
+              encode_base64( $req->portal . $req->env->{PATH_INFO},
                 '' );
         }
 
@@ -189,10 +189,6 @@ sub myNotifs {
 
     # Build template
     my $params = {
-        PORTAL        => $self->conf->{portal},
-        MAIN_LOGO     => $self->conf->{portalMainLogo},
-        SKIN          => $self->p->getSkin($req),
-        LANGS         => $self->conf->{showLanguages},
         NOTIFICATIONS => $_notifications,
         MSG           => $msg
     };
@@ -205,11 +201,11 @@ sub myNotifs {
 
 sub retrieveNotifs {
     my ( $self, $req ) = @_;
+    my $user = $req->userData->{ $self->conf->{whatToTrace} };
+    return [] unless $user;
 
     # Retrieve user's accepted notifications
-    $self->logger->debug( 'Searching for "'
-          . $req->userData->{ $self->conf->{whatToTrace} }
-          . '" accepted notification(s)' );
+    $self->logger->debug(qq{Searching for "$user" accepted notification(s)});
     my @_notifications = sort {
              $b->{epoch} <=> $a->{epoch}
           or $a->{reference} cmp $b->{reference}
